; ###############################################
; # pure OpenGL demo by d.j.peters (aka. Joshy) #
; ###############################################

EnableExplicit

XIncludeFile "GL.pbi"
XIncludeFile "GLU.pbi"
XIncludeFile "GLWindow.pbi"

XIncludeFile "car.pbi"

Enumeration
  #frmMain
  #glTarget
EndEnumeration  


Procedure frmResize(w.l,h.l)
  Protected flag.l
  If w<1: w=1:flag+1: EndIf
  If h<1: h=1:flag+1: EndIf  
  If flag<2
    glViewport(0, 0, w, h)
    glMatrixMode(#GL_PROJECTION)
    glLoadIdentity()
    gluPerspective(60.0, w/h, 0.1,1000.0)
    glMatrixMode(#GL_MODELVIEW)
  EndIf
EndProcedure

Procedure InitOpenGL()
  Protected Dim Light0Position.f(3), Dim Light0Diffuse.f(3)
  Protected Dim Light1Position.f(3), Dim Light1Diffuse.f(3), Dim Light1Specular.f(3)
  Protected Dim MatDiffuse.f(3),     Dim MatSpecular.f(3)
  
  
  Light0Position(0)=1:Light0Position(1)=5:Light0Position(2)=1:Light0Position(3)=0
  Light0Diffuse(0) =1:Light0Diffuse(1) =1:Light0Diffuse(2) =1:Light0Diffuse(3) =1
  
  Light1Position(0)=-2:Light1Position(1)=5.0:Light1Position(2)=0.0:Light1Position(3)=0
  Light1Diffuse(0) = 1:Light1Diffuse(1) =0.7:Light1Diffuse(2) =0.4:Light1Diffuse(3) =1
  Light1Specular(0)= 1:Light1Specular(1)=0.8:Light1Specular(2)=0.4:Light1Specular(3)=1
  
  MatDiffuse(0) =1:MatDiffuse(1) =1:MatDiffuse(2) =1:MatDiffuse(3) =1
  MatSpecular(0)=1:MatSpecular(1)=1:MatSpecular(2)=1:MatSpecular(3)=1
  
  ; quality
  glShadeModel(#GL_SMOOTH)
  glHint      (#GL_PERSPECTIVE_CORRECTION_HINT, #GL_NICEST) 
  
  ; pixel buffer
  glClearColor(0.1, 0.1, 0.2, 0.0)
  
  ; z buffer
  glEnable    (#GL_DEPTH_TEST)
  glClearDepth(1.0)
  
  ; blending (environment mapping)
  glBlendFunc (#GL_ONE, #GL_ONE)
  glDepthFunc (#GL_LEQUAL)
  glDisable   (#GL_BLEND)

  ; alpha channel (glass)
  glAlphaFunc (#GL_GREATER, 0.01)
  
  ; polygon visibility 
  glFrontface (#GL_CCW)
  glCullFace  (#GL_BACK)     
  ;glEnable    (#GL_CULL_FACE)
  glDisable(#GL_CULL_FACE)

  ; lights
  glEnable    (#GL_LIGHTING)
  glLightfv   (#GL_LIGHT0, #GL_POSITION,@Light0Position(0)) ; (1,5,1,0)
  glLightfv   (#GL_LIGHT0, #GL_DIFFUSE ,@Light0Diffuse(0))  ; (1,1,1,1)
  glEnable    (#GL_LIGHT0)

  glLightfv   (#GL_LIGHT1, #GL_POSITION,@Light1Position(0)) ; (-2,2,0,0)
  glLightfv   (#GL_LIGHT1, #GL_DIFFUSE ,@Light1Diffuse(0))  ; (1,0.7,0.4,1)
  glLightfv   (#GL_LIGHT1, #GL_SPECULAR,@Light1Specular(0)) ; (1,0.8,0.4,1)
  glEnable    (#GL_LIGHT1)

  ; material
  glEnable    (#GL_COLOR_MATERIAL)
  glMaterialfv(#GL_FRONT_AND_BACK, #GL_DIFFUSE ,@MatDiffuse(0))  ;(1,1,1,1)
  glMaterialfv(#GL_FRONT_AND_BACK, #GL_SPECULAR,@MatSpecular(0)) ;(1,1,1,1)
  glMaterialf (#GL_FRONT_AND_BACK, #GL_SHININESS,50)
  glMateriali (#GL_FRONT_AND_BACK, #GL_COLOR_MATERIAL_PARAMETER, #GL_SPECULAR)
  
  ; textures
  glTexGeni   (#GL_S, #GL_TEXTURE_GEN_MODE, #GL_SPHERE_MAP)
  glTexGeni   (#GL_T, #GL_TEXTURE_GEN_MODE, #GL_SPHERE_MAP)
  glTexGeni   (#GL_S, #GL_SPHERE_MAP, 0)
  glTexGeni   (#GL_T, #GL_SPHERE_MAP, 0)

  glDisable   (#GL_TEXTURE_GEN_S)
  glDisable   (#GL_TEXTURE_GEN_T)
  
  glEnable    (#GL_TEXTURE_2D)
EndProcedure  


;
; main
;
UsePNGImageDecoder()

Dim Cars.CAR_T(#MAX_CARS)

; no 3D engine, sprite or mouse
CompilerIf 0

If Not InitSprite()
  MessageRequester("error:","InitSprite()")
  End
EndIf
If Not InitKeyboard()
  MessageRequester("error:","InitKeyboard()")
  End
EndIf
If Not InitMouse()
  MessageRequester("error:","InitMouse()")
  End
EndIf

CompilerEndIf

If Not ExamineDesktops()
  MessageRequester("error:","ExamineDesktops()")
  End
EndIf

Define.l frmMain,scrMain,Event,Quit,EnvMap,Road,car
Define File.s
Define.l xRot=15,yRot,scr_w = DesktopWidth(0),scr_h=DesktopHeight(0)

#flags = #PB_Window_SizeGadget | #PB_Window_MinimizeGadget | #PB_Window_MaximizeGadget |  #PB_Window_ScreenCentered
If Not OpenGLWindow(#frmMain  , 0, 0, scr_w/2,scr_h/2, "OpenGL Car Rendering ...",#flags)
  MessageRequester("error:","OpenGLWindow()")
  End
EndIf

; i need this on my Linux box
; (may be a missing X11/GTK flush or sync)
Repeat : Delay(100) : Until IsWindow(frmMain)<>0

; setup OpenGL
InitOpenGL()

; load environmap
File = #EXT + "/Env." + #EXT
EnvMap = LoadGLTexture(File)
If EnvMap=0 
  MessageRequester("error:","LoadGLTexture() file: " + File)
  End
EndIf

; load road texture
File = #EXT + "/road_256x256." + #EXT
Road = LoadGLTexture(File)
If Road=0 
  MessageRequester("error:","LoadGLTexture() file: " + File)
  End
EndIf

; load cars and textures
For car=0 To #MAX_CARS-1
  Read.s File
  SetWindowTitle(#frmMain, "Load: " + File + " stand by ...")
  ReadCar(Cars(), car, File, EnvMap)
Next
car=0
SetWindowTitle(#frmMain, "render: " + Cars(car)\Name)


; loop while OpenGL window is open
Repeat
  Repeat
    Event = WindowEvent()
    Select Event
      Case #PB_Event_SizeWindow
        frmResize(WindowWidth(#frmMain),WindowHeight(#frmMain))  
      Case #PB_Event_CloseWindow
        Quit = 1
    EndSelect
  Until Event = 0 ; get all waiting events

  glClear(#GL_COLOR_BUFFER_BIT | #GL_DEPTH_BUFFER_BIT)
  glLoadIdentity()
  glTranslatef(0,0,-5)
  glRotatef(xRot,1,0,0)
  glRotatef(yRot,0,1,0)

  glEnable(#GL_TEXTURE_2D)
  glDisable(#GL_BLEND)
  glBindTexture(#GL_TEXTURE_2D,Road)
  glBegin(#GL_QUADS)
    glNormal3f(0,1,0)
    glColor3f(0.1,0.1,0.2)
    glTexCoord2f(0,0):glVertex3f(-7,0,-5)
    glTexCoord2f(0,1):glVertex3f(-7,0, 14)
    glTexCoord2f(1,1):glVertex3f( 3,0, 14)
    glTexCoord2f(1,0):glVertex3f( 3,0,-5)
  glEnd()     
  glDisable(#GL_TEXTURE_2D)
  
  glTranslatef(0,-Cars(car)\Models[0]\hdr\min[1],0)

  ; draw it 
  DrawCar(@Cars(car),0)
  ; show it
  SwapBuffers()
  yRot+1
  If yRot>=360
    yRot=0
    car=(car+1) % #MAX_CARS
   SetWindowTitle(#frmMain, "render: " + Cars(car)\Name)
  EndIf
  Delay(1000/60)
Until Quit = 1
; destroy render and device context
CloseGLWindow()
; thats all
End

DataSection
Data.s "BMWZ8"      ;   369 KB
Data.s "F1"         ;   394 KB
Data.s "Jaguar"     ;   403 KB
Data.s "Speedstr"   ;   407 KB
Data.s "LotusElise" ;   415 KB

Data.s "CL55"       ;   421 KB
Data.s "CL65"       ;   421 KB
Data.s "BMWM3"      ;   431 KB 
Data.s "Peugeot106" ;   541 KB

Data.s "Hummer"     ;  1122 KB

Data.s "Skyline"    ;  1241 KB
Data.s "Renault"    ;  1357 KB
Data.s "Prelude"    ;  1369 KB
Data.s "Golf"       ;  1372 KB
Data.s "AudiR8"     ;  1867 KB

Data.s "Carrera"    ;  2007 KB
Data.s "CaymanS"    ;  2057 KB
Data.s "Camaro"     ;  2266 KB
Data.s "TS50"       ;  2467 KB
Data.s "Mustang"    ;  2509 KB
EndDataSection




; IDE Options = PureBasic 4.60 (Windows - x86)
; CursorPosition = 219
; FirstLine = 187
; Folding = -
; EnableXP
; Executable = PureOpenGLDemo.exe
; CompileSourceDirectory