; CarTypes.pbi

XIncludeFile "GL.pbi"


#AABB  = 0 ; drawing AABB's for testing 

#MAX_CARS   = 20 
#MAX_MODELS =  3
#MAX_PARTS  = 100

; part classes
#UNK =   0 ; unknow
#FLW =   1 ; front left  wheel
#FRW =   2 ; front right wheel
#RRW =   4 ; rear  right wheel
#RLW =   8 ; rear  left  wheel
#CHA =  16 ; chassis
#GLA =  32 ; glass (window/light)
#ITR =  64 ; interiouer
#STW = 128 ; steering wheel
#MIS = 256 ; miscellaneous
#UND = 512 ; underside
#BRA =1024 ; part of brake
#WHEEL_MASK = $0F

; texture set 
#EXT = "png"

Structure HEADER_T
  Class.l
  Flags.l
  Name.s{16} ; 16 chars
  Min.f[3]
  Max.f[3]
  Pivot.f[4]
  nVertices.l
EndStructure

Structure PART_T
  hdr.HEADER_T 
  ; material
  Envmap.l
  Transparent.l
  TextureID.l         ; skin and wheel
  glList.l            ; precompiled DisplayList
EndStructure

Structure MODEL_T
  hdr.HEADER_T 
  nParts.l
  Parts.PART_T[#MAX_PARTS]
EndStructure

Structure CAR_T
  Name.s
  nModels.l
  Models.MODEL_T[#MAX_MODELS]
  Steering.f
  WheelRotation.f[4]
  Rotation.f[3]
  Position.f[3]
EndStructure


; load an image and create a OpenGL texture
Procedure LoadGLTexture(File.s)
  Protected ImgID.l  = LoadImage  (#PB_Any,File)
  If ImgID=0
    ProcedureReturn 0
  EndIf
  Protected.l x,y,index,color,glName,glFmt
  Protected Width.l  = ImageWidth (ImgID)
  Protected Height.l = ImageHeight(ImgID)
  Protected Depth.l  = ImageDepth (ImgID,#PB_Image_InternalDepth)/8
  Protected *pPixels = AllocateMemory(Width * Height * Depth + 4)
  Protected *pMem    = *pPixels  
  If Depth=3
    glFmt=#GL_RGB
  Else
    glFmt=#GL_RGBA
  EndIf
  ; create texture name and set properties
  glName = glGenTextures(1,@glName)
  glBindTexture(#GL_TEXTURE_2D,glName)
  glTexParameteri(#GL_TEXTURE_2D, #GL_TEXTURE_MIN_FILTER, #GL_NEAREST)
  glTexParameteri(#GL_TEXTURE_2D, #GL_TEXTURE_MAG_FILTER, #GL_NEAREST)

  ; MessageRequester("info:", "file: " + File + " glName: " + Str(glName) + " " + Str(Width) + " x " + Str(Height) + " x " + Str(Depth) )
  
  StartDrawing(ImageOutput(ImgID)) 
    Index=0
    For y=Height-1 To 0 Step -1
      For x=0 To Width-1
        color=Point(x,y)
        PokeL(*pMem,color)
        *pMem=*pMem+Depth
      Next
    Next
  StopDrawing()
  FreeImage(ImgID)
  
  ; set texture pixels
  glTexImage2D(#GL_TEXTURE_2D,0,Depth,Width,Height,0,glFmt,#GL_UNSIGNED_BYTE,*pPixels)

  ; free the pixel buffer
  FreeMemory(*pPixels)
 
  ; return GL name
  ProcedureReturn glName
EndProcedure

; a triangle fan car loader
Procedure.l ReadCar(Array cars.CAR_T(1), nCar.l, File.s, EnvMap.l)
  Protected.l m,p,v,i,j,Flag
  Protected Dim t.f(2),Dim v.f(3),Dim n.f(3)
  Protected Dim Textures.l(2)
  cars(nCar)\Name = File
  ; load textures from png or jpg folder
  Protected Skin.s  = #EXT + "/" + File + "Skin."  + #EXT
  Protected Wheel.s = #EXT + "/" + File + "Wheel." + #EXT      
             File    = "models/" + File + ".c3d"         
 
  If Not ReadFile(0,File)
    MessageRequester("error:","ReadCar() file: " + File)
    End
  EndIf
  
  ; load textures for the skin and wheels
  Textures(0)=LoadGLTexture(Skin)
  If Textures(0)=0 
    MessageRequester("error:","LoadGLTexture() file: " + Skin)
    CloseFile(0):End
  EndIf
  
  Textures(1)=LoadGLTexture(Wheel)
  If Textures(1)=0  
    MessageRequester("error:","LoadGLTexture() file: " + Wheel)
    CloseFile(0):End
  EndIf
  
  
  Cars(nCar)\nModels = ReadLong(0)
  If Cars(nCar)\nModels<1
    MessageRequester("error:","ReadCar(): wrong number of models !")
    CloseFile(0):End
  EndIf

  For m=0 To Cars(nCar)\nModels-1
    ; model header
    Cars(nCar)\Models[m]\Hdr\Class = ReadLong(0)
    Cars(nCar)\Models[m]\Hdr\Flags = ReadLong(0)
    File="":For i=0 To 15:File = File + Chr(ReadByte(0)):Next
    Cars(nCar)\Models[m]\Hdr\Name = File
    ; AABB box min and max values
    For i=0 To 2
      Cars(nCar)\Models[m]\Hdr\Min[i]  = ReadFloat(0)
    Next
    For i=0 To 2
      Cars(nCar)\Models[m]\Hdr\Max[i]  = ReadFloat(0)
    Next
    ; model pivot
    For i=0 To 2
      Cars(nCar)\Models[m]\Hdr\Pivot[i]= ReadFloat(0)
    Next
    ; model vertices from all parts
    Cars(nCar)\Models[m]\Hdr\nVertices = ReadLong(0)
    
    ; n parts
    Cars(nCar)\Models[m]\nParts = ReadLong(0)
    If Cars(nCar)\Models[m]\nParts<1
      MessageRequester("error:","ReadCar(): wrong number of parts !")
      CloseFile(0):End
    EndIf
    
    For p=0 To Cars(nCar)\Models[m]\nParts-1
      ; part header
      Cars(nCar)\Models[m]\Parts[p]\Hdr\Class = ReadLong(0)
      Cars(nCar)\Models[m]\Parts[p]\Hdr\Flags = ReadLong(0)
      File="":For i=0 To 15:File = File + Chr(ReadByte(0)):Next
      Cars(nCar)\Models[m]\Parts[p]\Hdr\Name=File
      ; AABB box
      For i=0 To 2
        Cars(nCar)\Models[m]\Parts[p]\Hdr\Min[i] = ReadFloat(0)
      Next
      For i=0 To 2
        Cars(nCar)\Models[m]\Parts[p]\Hdr\Max[i] = ReadFloat(0)
      Next
      ; model pivot
      For i=0 To 2
        Cars(nCar)\Models[m]\Parts[p]\Hdr\Pivot[i]= ReadFloat(0)
      Next                                                
      ; is it a wheel with it;s own local model space ?
      If (Cars(nCar)\Models[m]\Parts[p]\Hdr\Class & #WHEEL_MASK) 
        ; move AABBox from car space to it;s own wheel space
        For i=0 To 2
          Cars(nCar)\Models[m]\Parts[p]\Hdr\Min[i]-Cars(nCar)\Models[m]\Parts[p]\Hdr\Pivot[i]
        Next
        For i=0 To 2
          Cars(nCar)\Models[m]\Parts[p]\Hdr\Max[i]-Cars(nCar)\Models[m]\Parts[p]\Hdr\Pivot[i]
        Next
        
      EndIf

      ; part vertices
      Cars(nCar)\Models[m]\Parts[p]\Hdr\nVertices = ReadLong(0)
            
      ; read part material      
      
      ; environment texture ?
      Cars(nCar)\Models[m]\Parts[p]\EnvMap = ReadLong(0)
      If Cars(nCar)\Models[m]\Parts[p]\EnvMap<>0
        Cars(nCar)\Models[m]\Parts[p]\EnvMap=EnvMap
      EndIf
      
      ; glass ?
      Cars(nCar)\Models[m]\Parts[p]\Transparent = ReadLong(0)
      
      ; texture set 0 (skin/interior/glass or 1 wheel/brakes ?
      Cars(nCar)\Models[m]\Parts[p]\TextureID = ReadLong(0)-1
      If Cars(nCar)\Models[m]\Parts[p]\TextureID=0
        Cars(nCar)\Models[m]\Parts[p]\TextureID = Textures(0)
      ElseIf Cars(nCar)\Models[m]\Parts[p]\TextureID=1
        Cars(nCar)\Models[m]\Parts[p]\TextureID = Textures(1)
      EndIf    

      ; create a precompiled displaylist from part
      Cars(nCar)\Models[m]\Parts[p]\glList = glGenLists(1)
      glNewList(Cars(nCar)\Models[m]\Parts[p]\glList,#GL_COMPILE)

        glDisable(#GL_CULL_FACE)
        glBegin(#GL_TRIANGLE_STRIP)

        For v=0 To Cars(nCar)\Models[m]\Parts[p]\Hdr\nVertices-1
          t(0)=ReadFloat(0):t(1)=ReadFloat(0)
          n(0)=ReadFloat(0):n(1)=ReadFloat(0):n(2)=ReadFloat(0) 
          v(0)=ReadFloat(0):v(1)=ReadFloat(0):v(2)=ReadFloat(0)
          ; is it a wheel with it;s own center ?
          If (Cars(nCar)\Models[m]\Parts[p]\Hdr\Class & #WHEEL_MASK)
            ; move the vertices from car space to it;s own center wheel space
            For i=0 To 2 
              v(i)=v(i)-Cars(nCar)\Models[m]\Parts[p]\Hdr\Pivot[i]
            Next
          EndIf
          glTexCoord2fv(@t(0))
          glNormal3fv(@n(0))
          glVertex3fv(@v(0))
        
        Next ; next vertex
        
        glEnd() ; end of triangle strip
        glEnable(#GL_CULL_FACE)
      glEndList() ; end of compiled part
    
    Next ; next part
  
  Next ; next model
  
  CloseFile(0)

  ProcedureReturn 1
EndProcedure 

; draw a axis alinged boundary box
Procedure DrawAABB(*hdr.HEADER_T)
  glDisable(#GL_LIGHTING)
  glColor3f(0,0,1)
  
  glBegin(#GL_LINE_STRIP)
    glVertex3f(*hdr\Min[0],*hdr\Min[1],*hdr\Max[2])
    glVertex3f(*hdr\Min[0],*hdr\Max[1],*hdr\Max[2])
    glVertex3f(*hdr\Min[0],*hdr\Max[1],*hdr\Min[2])
    glVertex3f(*hdr\Min[0],*hdr\Min[1],*hdr\Min[2])
    glVertex3f(*hdr\Min[0],*hdr\Min[1],*hdr\Max[2])
  glEnd()

  glBegin(#GL_LINE_STRIP)
    glVertex3f(*hdr\Max[0],*hdr\Min[1],*hdr\Max[2])
    glVertex3f(*hdr\Max[0],*hdr\Max[1],*hdr\Max[2])
    glVertex3f(*hdr\Max[0],*hdr\Max[1],*hdr\Min[2])
    glVertex3f(*hdr\Max[0],*hdr\Min[1],*hdr\Min[2])
    glVertex3f(*hdr\Max[0],*hdr\Min[1],*hdr\Max[2])
  glEnd()
    
  glBegin(#GL_LINES)
    glVertex3f(*hdr\Min[0],*hdr\Min[1],*hdr\Max[2])
    glVertex3f(*hdr\Max[0],*hdr\Min[1],*hdr\Max[2])
    glVertex3f(*hdr\Min[0],*hdr\Max[1],*hdr\Max[2])
    glVertex3f(*hdr\Max[0],*hdr\Max[1],*hdr\Max[2])
    glVertex3f(*hdr\Min[0],*hdr\Max[1],*hdr\Min[2])
    glVertex3f(*hdr\Max[0],*hdr\Max[1],*hdr\Min[2])
    glVertex3f(*hdr\Min[0],*hdr\Min[1],*hdr\Min[2])
    glVertex3f(*hdr\Max[0],*hdr\Min[1],*hdr\Min[2])
  glEnd()
  
  glEnable(#GL_LIGHTING)
EndProcedure

; draw a part from precompiled OpenGL list.
Procedure DrawCarPart(*P.PART_T)
  If *P\TextureID>-1
    glBindTexture(#Gl_TEXTURE_2D,*P\TextureID)
    glEnable(#GL_TEXTURE_2D)
  Else
    glDisable(#GL_TEXTURE_2D)
  EndIf

  If *P\Transparent<>0
    glEnable  (#GL_BLEND)
    glColor3f (0,0,0)
    glCallList(*P\glList)
    glDisable (#GL_BLEND)
  Else
    glColor3f (1,1,1)
    glCallList(*P\glList)
  EndIf
  glDisable(#GL_TEXTURE_2D)

  If *P\EnvMap<>0
    glEnable(#GL_TEXTURE_GEN_S)
    glEnable(#GL_TEXTURE_GEN_T)
    glEnable(#GL_TEXTURE_2D)
    glEnable(#GL_BLEND)
    glBindTexture(#GL_TEXTURE_2D,*P\EnvMap)
    glcolor3f(0,0,0)
    glCallList(*P\glList)
    glDisable(#GL_BLEND)
    glDisable(#GL_TEXTURE_2D)
    glDisable(#GL_TEXTURE_GEN_S)
    glDisable(#GL_TEXTURE_GEN_T)
  EndIf
EndProcedure

; draw a model from a car
Procedure DrawCar(*Car.CAR_T,m.l)
  
  Protected p.l ; part counter
  
  If (#AABB<>0); car AABB
    DrawAABB(*Car\Models[m]\Hdr)
  EndIf   
   
  ; loop over all parts
  For p = 0 To *Car\Models[m]\nParts-1
    ; is it a wheel ?
    If (*Car\Models[m]\Parts[p]\hdr\class & #WHEEL_MASK)
      ; a wheel has it;s own local model space
      glPushMatrix() 
      glTranslatef(*Car\Models[m]\Parts[p]\Hdr\Pivot[0],*Car\Models[m]\Parts[p]\Hdr\Pivot[1],*Car\Models[m]\Parts[p]\Hdr\Pivot[2])
      ; is it a steering wheel ?
      If p<2 ; y axis (yaw)
        glRotatef(*Car\Steering,0,1,0)
      EndIf
      ; x axis (pitch)
      glRotatef(*Car\WheelRotation[p],-1,0,0)

      If (#AABB<>0) ; wheel AABB
        DrawAABB(*Car\Models[m]\Parts[p]\Hdr)
      EndIf
    EndIf
    
    ; finaly draw it 
    DrawCarPart(*Car\Models[m]\Parts[p])
    
    ; was it a wheel ?
    If (*Car\Models[m]\Parts[p]\hdr\class & #WHEEL_MASK)
      glPopMatrix()
    EndIf          

  Next ; next part  
  
EndProcedure

; IDE Options = PureBasic 4.60 (Windows - x86)
; CursorPosition = 5
; Folding = -
; EnableXP